/*************************************************************************************
*	irrlamb - http://irrlamb.googlecode.com
*	Copyright (C) 2010  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "graphics.h"
#include "save.h"
#include "globals.h"
#include "interface.h"
#include "input.h"
#include "config.h"
#include <ctime>

// Initializes the graphics system
int GraphicsClass::Init(int Width, int Height, bool Fullscreen, E_DRIVER_TYPE DriverType, IEventReceiver *EventReceiver) {
	ShowCursor = true;

	// irrlicht parameters
	SIrrlichtCreationParameters Parameters;
	Parameters.DriverType = DriverType;
	Parameters.Fullscreen = Fullscreen;
	Parameters.Bits = 32;
	Parameters.Vsync = 0;
	Parameters.Stencilbuffer = Config::Instance().Shadows;
	Parameters.AntiAlias = Config::Instance().AntiAliasing;
	Parameters.WindowSize.set(Width, Height);

	// Create the irrlicht device
	irrDevice = createDeviceEx(Parameters);
	if(irrDevice == NULL)
		return 0;

	irrDevice->setWindowCaption(L"irrlamb");
	irrDevice->setEventReceiver(EventReceiver);
	irrDevice->getCursorControl()->setVisible(false);

	// Save off global pointers
	irrDriver = irrDevice->getVideoDriver();
	irrScene = irrDevice->getSceneManager();
	irrGUI = irrDevice->getGUIEnvironment();
	irrFile = irrDevice->getFileSystem();
	irrTimer = irrDevice->getTimer();

	VideoModes.clear();

	// Generate a list of video modes
	IVideoModeList *VideoModeList = irrDevice->getVideoModeList();
	VideoModeStruct VideoMode;
	for(int i = 0; i < VideoModeList->getVideoModeCount(); i++) {
		VideoMode.Width = VideoModeList->getVideoModeResolution(i).Width;
		VideoMode.Height = VideoModeList->getVideoModeResolution(i).Height;
		VideoMode.BPP = VideoModeList->getVideoModeDepth(i);
		
		// Add the video mode
		if((VideoMode.BPP == 32 || VideoMode.BPP == 24) && VideoMode.Width >= 640) {
			VideoMode.String = stringw(VideoMode.Width) + stringw(" x ") + stringw(VideoMode.Height);
			VideoModes.push_back(VideoMode);
		}
	}

	//irrScene->setShadowColor(SColor(150, 0, 0, 0));

	DrawScene = true;
	ScreenshotRequested = false;

	return 1;
}

// Closes the graphics system
int GraphicsClass::Close() {
	
	// Close irrlicht
	irrDevice->drop();

	return 1;
}

// Erases the buffer and sets irrlicht up for the next frame
void GraphicsClass::BeginFrame() {
	irrDriver->beginScene(true, true, ClearColor);
	
	if(DrawScene)
		irrScene->drawAll();
}

// Draws the buffer to the screen
void GraphicsClass::EndFrame() {
	
	// Draw cursor
	if(ShowCursor)
		Interface::Instance().DrawImage(InterfaceClass::IMAGE_MOUSECURSOR, Input::Instance().GetMouseX(), Input::Instance().GetMouseY(), 16, 16);
	irrDriver->endScene();
	
	// Handle screenshots
	if(ScreenshotRequested) {
		ScreenshotRequested = 0;

		CreateScreenshot();
	}
}

// Returns the index of the current video mode
u32 GraphicsClass::GetCurrentVideoModeIndex() {

	// Find the video mode
	for(u32 i = 0; i < VideoModes.size(); i++) {
		if(Config::Instance().ScreenWidth == VideoModes[i].Width && Config::Instance().ScreenHeight == VideoModes[i].Height)
			return i;
	}

	return 0;
}

// Request screenshot
void GraphicsClass::SaveScreenshot() {
	ScreenshotRequested = 1;
}

// Create the screenshot
void GraphicsClass::CreateScreenshot() {

	// Get time
	time_t Now;
	time(&Now);

	// Get filename
	char Filename[32];
	strftime(Filename, 32, "%Y%m%d-%H%M%S.jpg", localtime(&Now)); 

	// Create image
	IImage *Image = irrDriver->createScreenShot();
	stringc FilePath;
	Save::Instance().GetScreenshotsPath(Filename, FilePath);
	irrDriver->writeImageToFile(Image, FilePath.c_str()); 
	Image->drop();
}
