/*************************************************************************************
*	irrlamb - http://irrlamb.googlecode.com
*	Copyright (C) 2010  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#ifndef OBJECT_H
#define OBJECT_H

// Libraries
#include <irrlicht/irrlicht.h>
#include <bullet/btBulletCollisionCommon.h>
#include <bullet/btBulletDynamicsCommon.h>

// Namespaces
using namespace irr;
using namespace core;
using namespace scene;
using namespace video;
using namespace io;
using namespace gui;

// Forward Declarations
struct SpawnStruct;
struct ConstraintStruct;

// Classes
class ObjectClass {

	public:

		enum ObjectType {
			NONE,
			PLAYER,
			ORB,
			COLLISION,
			SPHERE,
			BOX,
			CYLINDER,
			CONSTRAINT_HINGE,
			CONSTRAINT_D6,
			ZONE,
		};

		ObjectClass();
		virtual ~ObjectClass();

		// Updates
		virtual void Update(float FrameTime);
		void UpdateRender();
		void BeginFrame();
		virtual void EndFrame();
		
		// Replays
		virtual void UpdateReplay(float FrameTime);
		bool ReadyForReplayUpdate() const { return NeedsReplayPacket; }
		void WroteReplayPacket() { NeedsReplayPacket = false; }

		// Object properties
		void SetID(int Value) { ID = Value; }
		void SetTemplateID(int Value) { TemplateID = Value; }
		void SetDeleted(bool Value) { Deleted = Value; }
		void SetLifetime(float Value) { Lifetime = Timer + Value; }

		stringc GetName() const { return Name; }
		bool GetDeleted() const { return Deleted; }
		float GetLifetime() const { return Lifetime; }
		int GetType() const { return Type; }
		u16 GetID() const { return ID; }
		u16 GetTemplateID() const { return TemplateID; }

		// Rigid body
		bool IsHookable() const { return Hookable; }
		void Stop();
		
		void CalculateInterpolatedPosition(float BlendFactor);
		void SetPosition(const btVector3 &Position);
		const btVector3 &GetPosition() { return InterpolatedPosition; }
		const btVector3 &GetTruePosition() const { return RigidBody->getWorldTransform().getOrigin(); }

		void SetRotation(const btQuaternion &Rotation) { RigidBody->getWorldTransform().setRotation(Rotation); }
		btQuaternion GetRotation() { return RigidBody->getWorldTransform().getRotation(); }

		void SetLinearVelocity(const btVector3 &Velocity) { RigidBody->setLinearVelocity(Velocity); }
		const btVector3 &GetLinearVelocity() { return RigidBody->getLinearVelocity(); }

		void SetAngularVelocity(const btVector3 &Velocity) { RigidBody->setAngularVelocity(Velocity); }
		const btVector3 &GetAngularVelocity() { return RigidBody->getAngularVelocity(); }

		btRigidBody *GetBody() { return RigidBody; }
		ISceneNode *GetNode() { return Node; }

		virtual void HandleCollision(ObjectClass *OtherObject, const btPersistentManifold *ContactManifold, float NormalScale);
		bool IsTouchingGround() const { return TouchingGround; }

	protected:

		// Physics
		void CreateRigidBody(const SpawnStruct &Object, btCollisionShape *Shape);
		void SetProperties(const SpawnStruct &Object);
		void SetProperties(const ConstraintStruct &Object);

		// Attributes
		stringc Name;
		int Type;
		u16 ID, TemplateID;

		// State
		bool Deleted;

		// Life
		float Timer, Lifetime;
		
		// Physics and graphics
		ISceneNode *Node;
		btRigidBody *RigidBody;
		btVector3 LastPosition, InterpolatedPosition;
		btQuaternion LastRotation;

		// Replays
		bool NeedsReplayPacket;

		// Collision
		stringc CollisionCallback;
		bool TouchingGround, TouchingWall, Hookable;
};

#endif
