/*************************************************************************************
*	irrlamb - http://irrlamb.googlecode.com
*	Copyright (C) 2010  Alan Witkowski
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************************/
#include "player.h"
#include "../engine/globals.h"
#include "../engine/physics.h"
#include "../engine/objectmanager.h"
#include "../engine/input.h"
#include "../engine/config.h"
#include "../engine/level.h"
#include "sphere.h"
#include "constraint.h"
#include "springjoint.h"
#include "template.h"

// Constructor
PlayerClass::PlayerClass(const SpawnStruct &Object)
:	ObjectClass(),
	Camera(NULL),
	JumpTimer(0.0f),
	TorqueFactor(4.0f),
	GrappleState(GRAPPLE_READY),
	GrappleBody(NULL),
	GrappleRequestState(GRAPPLEREQUEST_NONE),
	StatGrappleLength(100.0f),
	GrappleEnabled(false) {

	// Graphics
	Node = irrScene->addSphereSceneNode(Object.Template->Radius, 24);
	Node->setMaterialTexture(0, irrDriver->getTexture("textures/player_outer0.png"));
	Node->setMaterialType(EMT_TRANSPARENT_ADD_COLOR);
	Node->setMaterialFlag(EMF_LIGHTING, false);
	Node->setMaterialType(EMT_ONETEXTURE_BLEND);
	Node->getMaterial(0).MaterialTypeParam = pack_texureBlendFunc(EBF_ONE, EBF_ONE);

	// Light
	//Light = irrScene->addLightSceneNode(0, core::vector3df(0.0f, 0.0f, 0.0f), video::SColorf(0.2f, 0.2f, 1.0f), 20.0f);

	// Add glow
	ISceneNode *InnerNode;
	InnerNode = irrScene->addBillboardSceneNode(Node, dimension2df(1.5f, 1.5f));
	InnerNode->setMaterialFlag(EMF_LIGHTING, false);
	InnerNode->setMaterialFlag(EMF_ZBUFFER, false);
	InnerNode->setMaterialTexture(0, irrDriver->getTexture("textures/player_glow0.png"));
	InnerNode->setMaterialType(EMT_ONETEXTURE_BLEND);
	InnerNode->getMaterial(0).MaterialTypeParam = pack_texureBlendFunc(EBF_ONE, EBF_ONE);

	// Particle system
	IParticleSystemSceneNode *ParticleSystem = irrScene->addParticleSystemSceneNode(false, Node);

	// Create the emitter
	IParticleEmitter *Emitter = ParticleSystem->createBoxEmitter(
		aabbox3df(-0.02f, -0.02f, -0.02f, 0.02f, 0.02f, 0.02f), 
		vector3df(0.0f, 0.0f, 0.0f),
		80, 80,
		SColor(0, 255, 255, 255), video::SColor(0, 255, 255, 255),
		1000, 1000);
	Emitter->setMinStartSize(dimension2d<f32>(0.25f, 0.25f));
	Emitter->setMaxStartSize(dimension2d<f32>(0.25f, 0.25f));
	ParticleSystem->setEmitter(Emitter);
	Emitter->drop();

	// Make the particles fade out
	IParticleAffector *ParticleAffector = ParticleSystem->createFadeOutParticleAffector();
	ParticleSystem->addAffector(ParticleAffector);
	ParticleAffector->drop();

	// Set materials
	ParticleSystem->setMaterialFlag(EMF_LIGHTING, false);
	ParticleSystem->setMaterialFlag(EMF_ZBUFFER, false);
	ParticleSystem->setMaterialTexture(0, irrDriver->getTexture("textures/player_trail0.png"));
	ParticleSystem->setMaterialType(EMT_TRANSPARENT_VERTEX_ALPHA);

	if(Physics::Instance().IsEnabled()) {

		// Create shape
		btSphereShape *Shape = new btSphereShape(Object.Template->Radius);

		// Set up physics
		CreateRigidBody(Object, Shape);
		RigidBody->setSleepingThresholds(0.1f, 0.1f);
	}

	SetProperties(Object);
	Hookable = false;
	if(CollisionCallback == "")
		CollisionCallback = "OnHitPlayer";
}

// Update the player
void PlayerClass::Update(float FrameTime) {
	ObjectClass::Update(FrameTime);
	//Light->setPosition(Node->getPosition());

	JumpTimer -= FrameTime;
	if(JumpTimer < 0.0f)
		JumpTimer = 0.0f;

	UpdateGrapple();
}

// Processes key presses
bool PlayerClass::ProcessKeyPress(int Key) {

	if(Key == Config::Instance().Keys[ACTIONS::JUMP]) {
		Jump();
	}	
	else
		return false;
	
	return true;
}

// Handles a mouse press
bool PlayerClass::HandleMousePress(int Button, int MouseX, int MouseY) {

	switch(Button) {
		// Right button
		case 1:
			if(GrappleEnabled)
				GrappleRequestState = GRAPPLEREQUEST_SHOOT;
		break;
		default:
			return false;
		break;
	}

	return true;
}

// Handles a mouse lift
bool PlayerClass::HandleMouseLift(int Button, int MouseX, int MouseY) {

	switch(Button) {
		// Right button
		case 1:
			if(GrappleEnabled)
				GrappleRequestState = GRAPPLEREQUEST_RELEASE;
		break;
		default:
			return false;
		break;
	}

	return true;
}

// Processes input from the keyboard
void PlayerClass::HandleInput() {
	
	// Get input direction
	vector3df Push(0.0f, 0.0f, 0.0f);
	if(Input::Instance().GetKeyState(Config::Instance().Keys[ACTIONS::MOVEFORWARD]))
		Push.Z = 1.0f;
	else if(Input::Instance().GetKeyState(Config::Instance().Keys[ACTIONS::MOVEBACK]))
		Push.Z = -1.0f;

	if(Input::Instance().GetKeyState(Config::Instance().Keys[ACTIONS::MOVELEFT]))
		Push.X = -1.0f;
	else if(Input::Instance().GetKeyState(Config::Instance().Keys[ACTIONS::MOVERIGHT]))
		Push.X = 1.0f;

	// Push the player
	if(!Push.equals(vector3df())) {
		Push.normalize();

		// Get push direction relative to camera
		matrix4 DirectionTransform;
		DirectionTransform.makeIdentity();
		DirectionTransform.setRotationDegrees(vector3df(0.0f, Camera->GetYaw(), 0.0f)); 
		DirectionTransform.transformVect(Push);

		// Apply torque
		vector3df RotationAxis = Push.crossProduct(vector3df(0.0f, -1.0f, 0.0f)) * TorqueFactor;
		RigidBody->activate();
		RigidBody->applyTorque(btVector3(RotationAxis.X, RotationAxis.Y, RotationAxis.Z));
	}
}

// Attempts to jump
void PlayerClass::Jump() {
	if(TouchingGround && JumpTimer == 0) {
		RigidBody->applyCentralImpulse(btVector3(0.0f, 5.0f, 0.0f));
		JumpTimer = 0.2f;
	}
}

// Stop the grapple and create a hook
void PlayerClass::GrappleTipHit(ObjectClass *Tip, ObjectClass *Object) {
	/*
	// Get template
	SpawnStruct GrappleFixed;
	int GrappleFixedIndex = Level::Instance().GetCustomObjectTemplateIndex(LevelClass::CUSTOM_GRAPPLEFIXED);
	Level::Instance().GetTemplateFromIndex(GrappleFixedIndex, GrappleFixed);

	// Create the fixed joint
	GrappleFixed.BodyA = Object;
	GrappleFixed.BodyB = Tip;
	GrappleJoints.push_back(ObjectManager::Instance().AddObject(new BasicJointClass(GrappleFixed)));
	Level::Instance().WriteCreateEvent(GrappleFixedIndex);

	// Get template
	SpawnStruct GrappleSpring;
	int GrappleSpringIndex = Level::Instance().GetCustomObjectTemplateIndex(LevelClass::CUSTOM_GRAPPLESPRING);
	Level::Instance().GetTemplateFromIndex(GrappleSpringIndex, GrappleSpring);

	// Create the spring joint
	GrappleSpring.BodyA = Tip;
	GrappleSpring.BodyB = this;
	GrappleJoints.push_back(ObjectManager::Instance().AddObject(new SpringJointClass(GrappleSpring)));
	Level::Instance().WriteCreateEvent(GrappleSpringIndex);

	GrappleState = GRAPPLE_HOOKED;
	*/
}

// Updates the grapple state
void PlayerClass::UpdateGrapple() {
/*
	switch(GrappleState) {
		case GRAPPLE_READY:
			if(GrappleRequestState == GRAPPLEREQUEST_SHOOT) {
				GrappleState = GRAPPLE_SHOOT;
				GrappleRequestState = GRAPPLEREQUEST_NONE;
			}
		break;
		case GRAPPLE_SHOOT: {

			SpawnStruct GrappleProperties;
			Level::Instance().GetTemplateFromIndex(Level::Instance().GetCustomObjectTemplateIndex(LevelClass::CUSTOM_GRAPPLETIP), GrappleProperties);

			// Set Position
			vector3df GrapplePosition(RigidBody->getWorldTransform().getOrigin()[0], RigidBody->getWorldTransform().getOrigin()[1], RigidBody->getWorldTransform().getOrigin()[2]);
			Level::Instance().SetDefaultPosition(GrapplePosition);

			// Create object
			GrappleBody = static_cast<GrappleClass *>(Level::Instance().CreateObjectFromTemplateIndex(Level::Instance().GetCustomObjectTemplateIndex(LevelClass::CUSTOM_GRAPPLETIP)));
			GrappleBody->SetGrappleParent(this);
			
			// Set direction
			btVector3 GrappleVelocity(Camera->GetDirection().X, Camera->GetDirection().Y, Camera->GetDirection().Z);
			GrappleVelocity *= 20.0f;
			GrappleBody->GetBody()->setLinearVelocity(GrappleVelocity + RigidBody->getLinearVelocity());

			GrappleState = GRAPPLE_FLYING;
		}
		break;
		case GRAPPLE_FLYING:
			if(GrappleRequestState == GRAPPLEREQUEST_RELEASE || !CheckGrappleDistance()) {
				GrappleState = GRAPPLE_DELETECHAIN;
				GrappleRequestState = GRAPPLEREQUEST_NONE;
			}
		break;
		case GRAPPLE_HOOKED:
			if(GrappleRequestState == GRAPPLEREQUEST_RELEASE) {
				GrappleState = GRAPPLE_DELETEJOINT;
				GrappleRequestState = GRAPPLEREQUEST_NONE;
			}
		break;
		case GRAPPLE_DELETEJOINT:

			// Remove the joint
			for(list<ObjectClass *>::Iterator Iterator = GrappleJoints.begin(); Iterator != GrappleJoints.end(); ++Iterator) {
				(*Iterator)->SetDeleted(true);
			}
			GrappleJoints.clear();

			GrappleState = GRAPPLE_DELETECHAIN;
		break;
		case GRAPPLE_DELETECHAIN:
			GrappleBody->SetDeleted(true);
			GrappleBody = NULL;

			GrappleState = GRAPPLE_READY;
		break;
	}
	*/
}

// Returns false if the grapple is too far away from the player
bool PlayerClass::CheckGrappleDistance() {

	//if(GrappleBody && (RigidBody->getWorldTransform().getOrigin() - GrappleBody->GetBody()->getWorldTransform().getOrigin()).length2() > StatGrappleLength)
	//	return false;

	return true;
}
